#!/usr/bin/env python
# 
# ddd.py
# Copyright (c) 2009 Neels J. Hofmeyr
#
#  This file is part of dddcms.
#
#  dddcms is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  dddcms is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with dddcms.  If not, see <http://www.gnu.org/licenses/>.

import os, sys, stat, shutil, time
import urllib, cgi
import random

from config import *

def read_file(path):
  try:
    f = open(path, "r")
    contents = f.read()
    f.close()
    return contents or ''
  except IOError:
    return None

def rewrite_file(path, content):
  try:
    file = open(path, 'w')
    file.write(content)
    file.close()
    return True
  except OSError:
    return False

def read_entries_file(path):
  contents = read_file(path)
  if not contents:
    return []
  return [line for line in contents.split(os.linesep) if len(line) > 0]

def write_entries_file(path, entries):
  print "WRITE ENTRIES", path, entries
  contents = os.linesep.join(entries) + os.linesep
  rewrite_file(path, contents)

def os_path_relpath(abs_path, base_path):
  "this replaces os.path.relpath which doesn't exist yet in my python."
  if abs_path.startswith(base_path):
    if base_path.endswith(os.path.sep):
      add = 0
    else:
      add = len(os.path.sep)
    return abs_path[ len(base_path)+add : ]

def mtime2str(mtime):
  if not mtime:
    return ''
  return time.strftime(DATE_FORMAT, time.localtime(mtime))

def mtime2filename(mtime=None):
  if not mtime:
    mtime = time.time()
  date_format = "%Y-%m-%d_%H-%M-%S"
  return time.strftime(date_format, time.localtime(mtime))

def mtime2path(mtime=None):
  if not mtime:
    mtime = time.time()
  date_format = '%Y' + os.path.sep + '%m' + os.path.sep + '%d_%H-%M-%S'
  return time.strftime(date_format, time.localtime(mtime))


if os.path.sep != '/':
  def path2url(path_str):
    return path_str.replace(os.path.sep, '/')

  def url2path(url_str):
    "this *only* replaces slashes with os.path.sep!"
    return url_str.replace('/', os.path.sep)
else:
  def path2url(path_str):
    return path_str

  def url2path(url_str):
    return url_str

def textarea2items(str):
  str = str.replace('\r', '\n')
  return [item for item in str.split('\n') if item]

class Path:
  "Everything that ddd does to filesystem paths."

  def __init__(my, path=None, base_path=None):
    my.base_path = ''
    my.abs_path = ''
    my.rel_elements = []
    my.rel_path = ''
    my.stat = None
    my.is_file = False
    my.is_dir = False
    
    # Note: don't use Path.annotation_cache, use DirNav.get_annotation(path)
    my.annotation_cache = None

    if isinstance(path, list) or isinstance(path, tuple):
      my.from_rel_elements(path, base_path)
    elif os.path.isabs(path):
      my.from_abs_path(path, base_path)
    else:
      if base_path is None:
        raise 'Path() constructor does not allow base_path == None'
      my.from_rel_path(path, base_path)

  def from_abs_path(my, abs_path, base_path):
    abs_path = os.path.abspath(abs_path)
    if base_path is None:
      base_path = abs_path

    my.base_path = base_path
    my.abs_path = abs_path
    my.rel_path = os_path_relpath(abs_path, base_path) or ''
    my.rel_elements = [el for el in my.rel_path.split(os.path.sep)
                       if len(el) > 0]

    if len(my.rel_elements) < 1:
      my.abs_path = my.base_path

    my.restat()
    return my
  
  def from_rel_path(my, rel_path, base_path):
    return my.from_abs_path( os.path.join(base_path, rel_path), base_path )

  def from_rel_elements(my, rel_elements, base_path):
    rel_elements = [el for el in rel_elements if el and len(el) > 0]
    return my.from_abs_path(os.path.join(base_path, *rel_elements),
                            base_path)

  def parent(my):
    return Path(my.parent_path(), my.base_path)

  def child(my, entry):
    child_path = my.child_path(entry)
    if child_path == my.abs_path:
      return my
    return Path(child_path, my.base_path)

  def sibling(my, entry):
    sibling_path = my.sibling_path(entry)
    if sibling_path == my.abs_path:
      return my
    return Path(sibling_path, my.base_path)

  def parent_path(my):
    return os.path.dirname(my.abs_path)

  def parent_rel_path_label(my):
    if len(my.rel_elements) <= 1:
      return ''
    return '/'.join(my.rel_elements[:-1]) + '/'

  def child_path(my, entry):
    if isinstance(entry, list) or isinstance(entry, tuple):
      if len(entry) < 1:
        return my.abs_path
      child = os.path.join(*entry)
    else:
      child = str(entry)
    return os.path.join(my.abs_path, child)

  def sibling_path(my, entry):
    if isinstance(entry, list) or isinstance(entry, tuple):
      if len(entry) < 1:
        return my.parent_path()
      child = os.path.join(*entry)
    else:
      child = str(entry)
    return os.path.join(my.parent_path(), child)

  def mkdir(my):
    "Create myself as a directory."
    if not os.path.exists(my.abs_path):
      os.makedirs(my.abs_path)

  def list(my):
    if not my.is_dir:
      return []
    return os.listdir(my.abs_path)

  def startswith(my, rel_paths_list):
    for path in rel_paths_list:
      if ( my.rel_path == path or
           my.rel_path.startswith(path) and 
           my.rel_path[len(path):].startswith(os.path.sep) ): 
        return True
    return False

  def endswith(my, suffixes_list, case_sensitive=True):
    if case_sensitive:
      return any(my.abs_path.endswith(suffix) for suffix in suffixes_list)
    name = my.abs_path.lower()
    return any(name.endswith(suffix.lower()) for suffix in suffixes_list)

  def make_parent_dirs(my):
    "create all path elements up to the parent of this path."
    parent_path = my.dirname()
    if not os.path.exists(parent_path):
      os.makedirs(parent_path)

  def touch(my):
    my.make_parent_dirs()
    file = open(my.abs_path, 'a')
    file.close()

  def rewrite(my, content):
    if not rewrite_file(my.abs_path, content):
      print "ERROR: failed to rewrite file", my.abs_path

  def rewrite_from(my, stream):
    try:
      f = open(my.abs_path, 'wb')
      while True:
        data = stream.read(1024*8)
        if not data:
          break
        f.write(data)
      f.close()
      return True
    except:
      return False


  def move_to_trash(my, trash_dir):
    destination = Path(my.rel_elements, trash_dir)
    print "MOVE", my.abs_path, "TO", destination.abs_path
    destination.make_parent_dirs()
    os.rename(my.abs_path, destination.abs_path)

  def really_delete(my):
    "CAREFUL! This deletes entire directory trees!"
    if my.is_dir:
      shutil.rmtree(my.abs_path)
    else:
      os.remove(my.abs_path)

  def rename(my, new_name, trash_dir):
    print 'aeh'
    print 'RENAME', my.abs_path, new_name, trash_dir
    if os.path.isabs(new_name):
      destination = Path(new_name.split(os.path.sep), my.base_path)
    else:
      destination = my.sibling(new_name)

    if destination.is_file or destination.is_dir:
      destination.move_to_trash(trash_dir)

    print 'RENAME', my.abs_path, destination.abs_path
    destination.make_parent_dirs()
    os.rename(my.abs_path, destination.abs_path)


  def __cmp__(my, path):
    if isinstance(path, Path):
      return cmp(my.abs_path, path.abs_path)
    return cmp(my.abs_path, path)

  def __str__(my):
    if my.is_file:
      t = 'f'
    elif my.is_dir:
      t = 'd'
    else:
      t = '?'
    return '(' + my.base_path + '//' + my.rel_path + ' ' + t + ')'

  def mtime(my):
    if not my.stat:
      return None
    return my.stat[stat.ST_MTIME]

  def time_str(my):
    return mtime2str(my.mtime())

  def dirname(my):
    return os.path.dirname(my.abs_path)

  def basename(my):
    if len(my.rel_elements) < 1:
      return ''
    return my.rel_elements[-1]

  def linkname(my):
    if my.is_dir:
      return my.basename() + '/'
    else:
      return my.basename()

  def linkname_urlencoded(my):
    return urllib.quote(my.linkname())

  def label(my, empty_if_root=False):
    if my.is_root():
      if empty_if_root:
        return ''
      else:
        return ROOT_LINK_LABEL
    return my.basename()


  def child_is_file(my, child):
    return os.path.isfile(my.child_path(child))

    # path = my.child_path(child)
    # try:
    #   mode = os.stat(path)[stat.ST_MODE]
    #   return stat.S_ISREG(mode) or stat.S_ISDIR(mode)
    # except OSError:
    #   return false

  def is_root(my):
    return len(my.rel_elements) < 1

  def bytes(my):
    if not my.stat:
      return None
    return my.stat[stat.ST_SIZE]

  def restat(my):
    my.is_file = False
    my.is_dir = False
    try:
      my.stat = os.stat(my.abs_path)
      mode = my.stat[stat.ST_MODE]
      if stat.S_ISREG(mode):
        my.is_file = True
      elif stat.S_ISDIR(mode):
        my.is_dir = True
    except OSError:
      my.stat = None

  def exists(my):
    return my.stat is not None

  def abs_url(my, root_url):
    "constructs an absolute url from root_url and my.rel_elements"
    if not root_url.endswith('/'):
      root_url += '/'
    url = root_url + '/'.join(my.rel_elements)
    if my.is_dir and not url.endswith('/'):
      url += '/'
    return urllib.quote(url)

  def link_from(my, otherdir, root_url):
    if my.abs_path.startswith(otherdir.abs_path):
      return urllib.quote(
               '/'.join(my.rel_elements[len(otherdir.rel_elements):]) )
    else:
      return my.abs_url(root_url)

  def size_str(my):
    "used from the 'files' template"
    bytes = my.bytes() or 0

    i = 1
    end = len(_path_size_units)
    while bytes > _path_size_units[i][0] and i < end:
      i += 1
    i -= 1
    if i == 0:
      return str(bytes) + _path_size_units[0][1]

    factor, unit = _path_size_units[i]
    return str(round(float(bytes) / factor, 1)) + unit

_path_size_units = [
                    (1, ' b'),
                    (1<<10L, ' kb'),
                    (1<<20L, ' Mb'), 
                    (1<<30L, ' Gb'), 
                    (1<<40L, ' Tb'), 
                    (1<<50L, ' Pb')]



class Paths:
  "some useful functions on a list of Path instances."
  def __init__(my, paths):
    my.paths = paths

  def newest_mtime(my, default=None, files_only=True):
    
    def newest(a, b):
      if a.mtime() > b.mtime():
        return a
      return b

    paths = my.paths
    if len(paths) < 1:
      return default
    
    if files_only:
      paths = [path for path in paths if path.is_file]
    if len(paths) < 1:
      return default
    
    r = reduce(newest, paths)
    mtime = r.mtime()
    return mtime

  def find_and_remove(my, basename):
    "If item exists, remove it and return True."
    for path in my.paths:
      if basename == path.basename():
        my.paths.remove(path)
        return True
    return False

  def filter(my, list_of_strings):
    my.paths = [path for path in my.paths
                if path.basename() in list_of_strings]


class Photo:
  "Keeps track of an image file and its low-resolution representations."
  image_size_cache = {}

  def __init__(my, ddd, dir, file):
    my.dir = dir
    my.file = file
    my.ddd = ddd

  def label(my):
    return my.file.label()

  def size_str(my):
    return my.file.size_str()

  def url(my):
    "returns the url relative to its navigation parent dir."
    return my.file.link_from(my.dir, my.ddd.root_url)

  def lores_url(my):
    "returns the complete url that gives a possibly scaled view of the photo"
    if my.file.startswith( (my.ddd.thumbs_dir.abs_path,
                            my.ddd.lo_res_dir.abs_path) ):
      # don't make cached images of cached images
      # (in case a user browses the image cache online)
      return my.url()
    return my._scaled_url(my.ddd.lo_res_dir, LORES_SIZE)

  def thumb_url(my):
    "makes a thumb if necessary and returns its complete url"
    return my._scaled_url(my.ddd.thumbs_dir, THUMBS_SIZE)

  def view_url(my):
    # This one is different -- it avoids entering a "_photos"
    # subfolder. The counterpart to this link is in
    # DddRequestHandler.probe_request_and_list_nav().
    return my.file.linkname_urlencoded() + '?' + PHOTO_VIEW_ARG

  def _scaled_url(my, cache_dir, size):
    # don't make cached images of cached images or the trash.
    # (in case a user browses the image cache or trash online)
    if my.ddd.is_image_cache(my.file):
      return my.url()

    if not PREVIEW_IMAGES_IN_TRASH and my.ddd.is_trash(my.file):
      return my.url()

    scaled = Path(os.path.join(cache_dir.rel_path, my.file.rel_path + '.jpg'),
                  my.file.base_path)

    if not scaled.is_file or scaled.mtime() <= my.file.mtime():
      # create or update the scaled image
      try:
        want = [float(x) for x in size.split('x')]

        import PythonMagick
        print 'READING IMAGE', my.file.abs_path
        img = PythonMagick.Image(my.file.abs_path)
        got = [img.size().width(), img.size().height()]

        if all(got[i] <= want[i] for i in range(2)):
          # create an empty placeholder, remembering the timestamp.
          print 'IMAGE KEPT UNSCALED', my.file.abs_path, 'SINCE IT FITS', size
          scaled.touch()
          return my.file.link_fromurl()
        
        print 'SCALING IMAGE', my.file.abs_path, 'TO FIT', size
        img.sample(size)
        scaled.make_parent_dirs()
        img.write(scaled.abs_path)
        print "IMAGE WRITTEN:", scaled.rel_path
      except:
        return my.url()

    if scaled.bytes() == 0:
      return my.url()

    return my.ddd.root_url + path2url(scaled.rel_path)



class DirNav:
  """Given a db path, this class figures out which children are visible
  in the navigation."""

  def __init__(my, requesthandler, dir=None,
               nav_only=False, do_photo_view=False):
    "DIR must be an instance of the Path class, and it must be a valid dir."

    if not dir:
      dir = requesthandler.dir

    my.ddd = requesthandler.ddd
    my.requesthandler = requesthandler
    my.dir = dir
    my.refresh(nav_only, do_photo_view)


  def refresh(my, nav_only=False, do_photo_view=False):
    dir = my.dir

    my.do_photo_view = do_photo_view
    my.title = dir.label(empty_if_root=False)
    my.mtime = None
    my.text = ""
    my.nav_children = []
    my.files = []
    my.photos = []
    my.annotated_names_cache = None

    children = my.get_sorted_children()

    if not nav_only:
      my.mtime = children.newest_mtime(default=dir.mtime(), files_only=True)

    my.items_hidden = []
    if my.has_items_hidden:
      my.items_hidden = read_entries_file(dir.child_path(ITEMS_HIDDEN_NAME))
    
    do_photos = not my.has_photos_dir

    for child in children.paths:
      if child.basename() in my.items_hidden:
        continue
      if child.is_file:
        if nav_only:
          continue
        if do_photos and child.endswith(PHOTO_SUFFIXES,
                                        PHOTO_SUFFIXES_ARE_CASE_SENSITIVE):
          my.photos.append(Photo(my.ddd, my.dir, child))
        elif not my.has_files_dir:
          my.files.append(child)
      elif child.is_dir:
        my.nav_children.append(child)

    if my.has_photos_dir and not nav_only:
      photo_paths = my.get_sorted_children(
                      dir.child(PHOTOS_DIR_NAME)
                    ).paths
      for photo_path in photo_paths:
        if photo_path.endswith(PHOTO_SUFFIXES,
                               PHOTO_SUFFIXES_ARE_CASE_SENSITIVE):
          my.photos.append(Photo(my.ddd, my.dir, photo_path))

    if not my.photos:
      my.do_photo_view = False

    if my.has_text_file and not (my.do_photo_view or nav_only):
      my.text = read_file(dir.child_path(TEXT_CONTENT_NAME))

    if my.has_files_dir and not (my.do_photo_view or nav_only):
      my.files = my.get_sorted_children(
                      dir.child(FILES_DIR_NAME)
                    ).paths

    if my.photos and not my.files and not my.text:
      my.do_photo_view = True


  def annotated_names(my):
    if not my.has_annotations:
      return []
      
    if my.annotated_names_cache is not None:
      return my.annotated_names_cache

    annotations_dir = my.dir.child(ANNOTATIONS_DIR_NAME)
    my.annotated_names_cache = annotations_dir.list()
    return my.annotated_names_cache
      
  def get_annotation_name(my, path):
    return path.basename() + ANNOTATION_SUFFIX

  def get_annotation_path(my, annotation_name):
    return my.dir.child_path([ANNOTATIONS_DIR_NAME, annotation_name])

  def get_annotation_path_instance(my, path):
    return Path(
             my.get_annotation_path(my.get_annotation_name(path)),
             my.ddd.db_root
           )

  def get_annotation(my, path):
    """If necessary reads the annotation and returns it. The annotation
    is cached in the given Path instance and is not read again if the
    path.annotation_cache is already not None."""
    if path.annotation_cache:
      return path.annotation_cache

    annotation_name = my.get_annotation_name(path)
    if annotation_name not in my.annotated_names():
      path.annotation_cache = ''
      return path.annotation_cache

    path.annotation_cache = read_file(my.get_annotation_path(annotation_name))
    return path.annotation_cache

  def set_annotation(my, path, annotation, trash_dir):
    annotation_path = my.get_annotation_path_instance(path)
    if not annotation_path.startswith(DB_EXLUDE):
      if not annotation:
        if annotation_path.is_file:
          print 'REMOVING ANNOTATION', annotation_path, annotation
          try:
            annotation_path.move_to_trash(trash_dir)
          except:
            pass
          return
      else:
        print 'WRITING ANNOTATION', annotation_path, annotation
        annotation_path.make_parent_dirs()
        annotation_path.rewrite(annotation)
        path.annotation_cache = annotation
      # reload if needed
      my.annotated_names_cache = None


  def get_sorted_children(my, dir=None):
    """Reads my.dir's file listing and interprets the ddd control file names.
    With argument DIR passed non-null, this reads the files from the
    '_files' directory, and expects the passed DIR to point at '_files'.
    If DIR is supplied, this does not impact the instance. Otherwise,
    control file names get detected and removed from the listing."""
    if not dir:
      dir = my.dir
      is_files_dir = False
    else:
      is_files_dir = True

    # see if any sort file is given. If yes, we list those
    # first (for sort order, only). 

    sorted = [dir.child(entry)
              for entry in read_entries_file(my.dir.child_path(SORT_NAME))
             ]

    unsorted = [dir.child(entry) for entry in dir.list()
                if all( sorted_entry.basename() != entry
                        for sorted_entry in sorted )
               ]

    unsorted.sort(ITEMS_CMP)
    sorted.extend( unsorted )

    children = Paths([ path for path in sorted
                       if (path.is_file or path.is_dir)
                          and not path.startswith(DB_EXLUDE)
                     ])

    if not is_files_dir:
      children.find_and_remove(SORT_NAME)
      my.has_text_file = children.find_and_remove(TEXT_CONTENT_NAME)
      my.has_files_dir = children.find_and_remove(FILES_DIR_NAME)
      my.has_annotations = children.find_and_remove(ANNOTATIONS_DIR_NAME)
      my.has_photos_dir = children.find_and_remove(PHOTOS_DIR_NAME)
      my.has_items_hidden = children.find_and_remove(ITEMS_HIDDEN_NAME)

    return children


  def write_entries_file(my, file, paths):
    write_entries_file(file.abs_path,
                       [path.basename() for path in paths])


class Dialog:
  """keeps track of a user interaction. That's either one that requires
  entering data or one that displays a message from ddd to the user."""

  def __init__(my, before_nav_handler, after_nav_handler,
               title, items, data=None, id=None):
    if not id:
      id = str(random.random())
    my.id = id
    my.title = title
    my.items = items
    if before_nav_handler:
      my.before_nav_handler_str = before_nav_handler.__name__
    else:
      my.before_nav_handler_str = None
    if after_nav_handler:
      my.after_nav_handler_str = after_nav_handler.__name__
    else:
      my.after_nav_handler_str = None
    my.data = data

  def has_handler(my):
    return not (my.before_nav_handler_str is None and
                my.after_nav_handler_str is None)

  def get_handler(my):
    return DialogHandler(my.id, my.before_nav_handler_str,
                         my.after_nav_handler_str, my.data)


class DialogHandler:
  def __init__(my, id, before_nav_handler_str, after_nav_handler_str,
               data):
    my.id = id
    my.before_nav_handler_str = before_nav_handler_str
    my.after_nav_handler_str = after_nav_handler_str
    my.data = data

  def handle(my, dddrequesthandler, args, is_before_nav):
    if is_before_nav:
      handler_str = my.before_nav_handler_str
    else:
      handler_str = my.after_nav_handler_str
    if not handler_str or handler_str == '':
      return None

    id = args.get(DIALOG_ID_NAME)
    if not id or id != my.id:
      return None

    handler = getattr(dddrequesthandler, handler_str)
    return handler(args, my.data)


# Dialog items ###################################
# Make dictionary representations of the HTML form elements for Dialog().
# They are made into HTML code in the template .../templates/dialog .

def to_dict(**args):
  "tiny helper"
  return args

def di_text(text, style='text', span=None, markup=False):
  if not markup:
    text = cgi.escape(text)
  return to_dict(text=text, style=style, span=span, label=None)

def _di_entry(type, name, label=None, prefill='', style='entry', markup=False):
  if label and not markup:
    label = cgi.escape(label)
  dict = to_dict(name=name, label=label, style=style,
                 prefill=cgi.escape(prefill).replace('"', '&quot;'))
  dict[type] = True
  return dict

def di_entry(name, label=None, prefill='', style='entry', markup=False):
  return _di_entry('entry', name, label, prefill, style, markup)

def di_password(name, label=None, prefill=''):
  "if label is None, the name will be taken as label"
  return _di_entry('password', name, label, prefill)

def di_textarea(name, label=None, prefill='', style='textarea', tinymce=None,
                span=None):
  if prefill:
    prefill = cgi.escape(prefill)

  return to_dict(textarea=True, name=name, label=label, prefill=prefill,
                 style=style, tinymce=tinymce, span=span)

def di_item_list(name, label=None, items=[], style='item_list'):
  prefill = '\n'.join(items)
  return di_textarea(name, label, prefill, style)

def di_checkbox(name, label, checked=False):
  return to_dict(checkbox=True, name=name, label=label, checked=checked)

def di_radio(name, value, selected=False, radio_label=None):
  return to_dict(radio=True, name=name, value=value,
                 radio_label=radio_label, selected=selected)

def di_upload(name, label=None):
  return to_dict(upload=True, name=name, label=label)

def di_submit(label='Go'):
  return to_dict(submit=True, submit_label=label)

def di_submit_choices(names_and_labels, mark_name=None,
                      style='submit_choices'):
  return to_dict(submit_choices=True, mark_name=mark_name, style=style,
                 names_and_labels=names_and_labels)

def di_hidden(name, value):
  "Don't use this, use Dialog(..., data=('your', 'data')) instead."
  return to_dict(hidden=True, name=name, value=value)

def di_hr(span=None):
  return to_dict(hr=True, span=span)

def di_start_table(*headings):
  return to_dict(start_table=True, headings=headings)

def di_end_table(*row_items):
  return to_dict(end_table=True)

def di_link(url, label):
  return to_dict(link=True, url=url, link_label=label)

def di_empty():
  return to_dict(empty=True)

##################################################


class DddRequestHandler:
  """One of these is created for each request. This is "everything"."""
  
  # Dialogs ########################################
  # The dialogs (messages or forms) ddd sends to the user.

  def dialog_error(my, msg):
    return Dialog(None, None, 'fail', [ di_text(msg, markup=True) ])

  def dialog_login(my):
    return Dialog(my.dialog_login_handler, None, 'login',
                  [ di_entry(DIALOG_LOGIN_NAME_ARG, 'name:'),
                    di_password(DIALOG_LOGIN_PASS_ARG, 'password:'),
                    (di_empty(), di_submit('login')) ] )

  def dialog_login_handler(my, results, data):
    login_name = results.get(DIALOG_LOGIN_NAME_ARG)
    login_pass = results.get(DIALOG_LOGIN_PASS_ARG)
    print "LOGIN NAME", login_name
    if login_pass and login_pass == USERS.get(login_name):
      my.s.login = login_name
      # user is now logged in.
    return None


      
  def dialog_text(my):
    dir = my.nav.dir

    prefill = my.nav.text or ''

    return Dialog(my.dialog_edit_text_handler, my.dialog_tab_choice,
                  dir.label(),
                  [ di_textarea(DIALOG_EDIT_TEXT, prefill=prefill, tinymce=1,
                                span=2),
                    (di_submit('Save'),
                     di_text('hint: use the fullscreen button',
                             style='hint')),
                    di_hr(2),
                    my._dialog__make_tabs(my.dialog_text),
                  ],
                  (dir, prefill))

  def dialog_edit_text_handler(my, results, data):
    dir, prefill = data

    if not dir == my.dir:
      return my.dialog_error(
                  "Failure -- the edit was sent to the wrong place??")

    # make sure any text changes are saved.
    text = results.get(DIALOG_EDIT_TEXT) or ''
    if text == prefill:
      # nothing changed.
      return None

    text_file = dir.child(TEXT_CONTENT_NAME)

    if text_file.is_dir:
      return my.dialog_error("can't edit, "
                             + TEXT_CONTENT_NAME + "is a folder: "
                             + path2url(text_file.rel_path))

    if text_file.is_file:
      # has anything changed?
      current_text = read_file(text_file.abs_path)
      if current_text == text:
        # same as on disk. don't replace.
        return None

      if current_text != prefill:
        # darn it, someone else has edited the text in a conflicting way!
        return my.dialog_error(
"""Oh no, someone else has edited the same text in the meantime!<br>
You need to
<ul>
<li>go back (click your web browser's back button),</li>
<li>copy your text,</li>
<li>open a new browser window on the text editing page,</li>
<li>see what else was edited, and</li>
<li>merge your edits with the other edits, manually.</li>
</ul>
Sorry for the inconvenience. If you want this handled
intelligently, write to <a href="mailto:neels@hofmeyr.de">neels</a>.""")

      if current_text == text:
        # nope. don't replace.
        return None

      # move away the previous text.
      text_file.move_to_trash(my.trash_dir())

    if len(text) > 0:
      text_file.rewrite(text)

    return None


  dialog_tabs = (
             ('dialog_text', 'text'),
             ('dialog_config', 'config'),
             ('dialog_add', 'add'),
             ('dialog_sort', 'sort'),
             ('dialog_remove', 'delete'),
             ('dialog_move', 'move'),
             ('dialog_copy', 'copy'),
             )

  def _dialog__make_tabs(my, current, di=None):
    choices = di_submit_choices(my.dialog_tabs, current.__name__, 'item_tabs')
    if di is None:
      return [choices]
    elif not di:
      di = [di_text('nothing', style='hint')]

    return [choices,
            di_hr(),
            di_start_table(),
        ] + di + [
            di_end_table()
           ]


  def dialog_last_tab_choice(my, choice=None):
    dir = my.nav.dir

    if not choice:
      choice_name = my.s.last_dialog_tab
      if choice_name:
        choice = getattr(my, choice_name)
      if choice == my.dialog_text or not choice:
        choice = my.dialog_config

    my.s.last_dialog_tab = choice.__name__

    return choice()

  def dialog_tab_choice(my, results, data):
    "handles the choice result after each of the item tabs are submitted."
    for name, label in my.dialog_tabs:
      if name in results:
        my.s.last_dialog_tab = name
        func = getattr(my, name)
        return func()
    return None
    
  def dialog_add(my):
    dir = my.nav.dir
    
    return Dialog(my.dialog_add_handler, my.dialog_tab_choice,
                  dir.label(),
                  my._dialog__make_tabs(
                    my.dialog_add,
                    [ di_item_list(DIALOG_ADD_NAME_ARG, 'add folder(s)',
                                   style='add_list'),
                      (di_empty(), di_submit('Add')),
                      di_hr(2),
                      di_upload('upload_file', 'upload file'),
                      (di_empty(), di_submit('Add')),
                    ]),
                  dir)

  def dialog_add_handler(my, results, data):
    dir = data
    errors = []
    
    file = results.get('upload_file')
    if file is not None and file.filename:
      # receive the uploaded file
      new_file = dir.child(file.filename)
      if new_file.startswith(DB_EXLUDE) or (
         new_file.exists() and not new_file.is_file):
        errors.append(file.filename + ' (upload is lost!)')
      else:
        if new_file.is_file:
          new_file.move_to_trash(my.trash_dir())
        print 'RECEIVING FILE', new_file
        if not new_file.rewrite_from(file.file):
          print 'RECEIVING FILE FAILED'
          errors.append(file.filename + ' (upload is lost!)')

    add_names = textarea2items(results.get(DIALOG_ADD_NAME_ARG) or '')
    first = None
    for add_name in add_names:
      new_dir = dir.child(add_name)
      try:
        new_dir.mkdir()
        if not first:
          first = new_dir
      except:
        errors.append(add_name)

    if errors:
      return my.dialog_error('unable to create:<br>' +
                             '<br>'.join(errors))

    if len(add_names) == 1 and first:
      # add of a single item succeeded, redirect to the new URL
      my.ddd.ddd.redirect(first)
      # not reached

    return None


  def _dialog_remove__make_file_list(my, deletables,
                                     hidden_items_at=None,
                                     invisible_items_at=None,
                                     checked=False):
    di = []
    nr = 0
    for deletable in deletables:
      if nr == hidden_items_at:
        text = 'delete hidden items:'
      elif nr == invisible_items_at:
        text = 'delete invisible items:'
      else:
        text = None
      if text:
        if nr > 0:
          di.append( (di_empty(), di_submit(DELETE_BUTTON_LABEL)) )
          di.append( di_hr(2) )
        di.append( di_text(text, style='hint') )

      di.append( di_checkbox('remove_' + str(nr),
                             cgi.escape(deletable.linkname()),
                             checked) )
      nr += 1
    return di


  def dialog_remove(my):
    def found(name, deletables):
      return any(d.basename() == name for d in deletables)

    # show a list of removable children
    dir = my.nav.dir

    # list nav items at the top
    deletables = my.nav.nav_children + my.nav.files

    for photo in my.nav.photos:
      deletables.append(photo.file)

    more_deletables = []

    items_hidden = my.nav.items_hidden

    for name in items_hidden:
      if found(name, deletables):
        continue
      child = dir.child(name)
      if child.startswith(DB_EXLUDE):
        continue

      if child.is_file or child.is_dir:
        more_deletables.append(child)


    if more_deletables:
      hidden_items_at = len(deletables)
      deletables.extend(more_deletables)
    else:
      hidden_items_at = None

    # also add files and photos that are invisible but not hidden
    # explicitly.
    more_deletables = []
    dirs = []
    listing = dir.list()
    for name in listing:
      if found(name, deletables):
        continue
      child = dir.child(name)
      if child.startswith(DB_EXLUDE):
        continue
      if child.is_file:
        more_deletables.append(child)
      elif child.is_dir:
        dirs.append(child)
    more_deletables.extend(dirs)

    if more_deletables:
      invisible_items_at = len(deletables)
      deletables.extend(more_deletables)
    else:
      invisible_items_at = None


    di = my._dialog_remove__make_file_list(
              deletables, hidden_items_at, invisible_items_at)


    if di:
      # show a submit button every N steps down.
      submit_step = 10
      added = 0
      for i in range(int(len(di) / submit_step)):
        pos = (i+1) * submit_step

        # don't show another submit button for too few items.
        if (hidden_items_at is not None and
            abs(pos - hidden_items_at) < submit_step / 2):
          continue

        if (invisible_items_at is not None and
            abs(pos - invisible_items_at) < submit_step / 2):
          continue

        if abs(pos - len(di)) < submit_step / 2:
          break

        # with each added item, the list items below are shifted down
        # by one.
        di.insert(pos + added, (di_empty(), di_submit(DELETE_BUTTON_LABEL)) )
        added += 1

      di.append( (di_empty(), di_submit(DELETE_BUTTON_LABEL)) )

    return Dialog(None, my.dialog_remove_handler,
                  dir.label(),
                  my._dialog__make_tabs(
                    my.dialog_remove,
                    di),
                  (deletables, False))

  def dialog_remove_handler(my, results, data):
    deletables, really_delete = data

    if really_delete:
      # already asked for confirmation. What was the answer?
      rr = results.get('yes_really_remove')
      if not rr:
        # aborted delete.
        return my.dialog_tab_choice(results, data)

    really_deletables = []
    errors = []

    do_refresh = False

    n = 0
    for deletable in deletables:
      # is it checked?
      if results.get('remove_' + str(n)):
        print 'GOT remove_' + str(n), results.get('remove_' + str(n))
        # if in any of the automated directories, allow real remove.
        if my.ddd.is_trash_or_image_cache(deletable):
          really_deletables.append(deletable)
        elif not deletable.startswith(DB_EXLUDE):
          do_refresh = True
          try:
            deletable.move_to_trash(my.trash_dir())
          except:
            errors.append(deletable.rel_path)

          # also remove the annotation, if any.
          my.nav.set_annotation(deletable, None, my.trash_dir())

      n += 1 

    print "REALLY DELETABLES", really_deletables

    if len(really_deletables) > 0:
      if really_delete:
        do_refresh = True
        # delete!
        for deletable in really_deletables:
          try:
            deletable.really_delete()
          except:
            errors.append(deletable.rel_path)
      else:
        # ask for real deletion confirmation
        deletables = really_deletables
        di = my._dialog_remove__make_file_list(deletables, checked=True)
        di.append( (di_empty(), di_submit_choices(
                                  (('dont_remove', 'No'),
                                   ('yes_really_remove', 'Yes'))
                                                 )
                   )
                 )
        return Dialog(None, my.dialog_remove_handler,
                      'PERMANENTLY DELETE??', di,
                      (deletables, True))
        # Note: the next dialog again checks each and every path whether
        #       it is ok to be really deleted. That's safer.
      

    if do_refresh:
      my.nav.refresh()

    if len(errors) > 0:
      return my.dialog_error('Failed to remove:<br>' +
                             '<br>'.join(errors))
    return my.dialog_tab_choice(results, data)


  def dialog_sort(my, fill=None):
    dir = my.dir

    items = read_entries_file( dir.child_path(SORT_NAME) )

    if fill is not None:
      show_items = fill
      show_items.extend([ item.basename() for item in my.nav.nav_children
                          if item.basename() not in show_items ])
      
      more = [ item.file.basename() for item in my.nav.photos
               if item.file.basename() not in show_items ]
      if more and show_items and show_items[-1] != '---':
        show_items.append('---')
      show_items.extend(more)

      more = [ item.basename() for item in my.nav.files
               if item.basename() not in show_items ]
      if more and show_items and show_items[-1] != '---':
        show_items.append('---')
      show_items.extend(more)
    else:
      # copy
      show_items = items + []
      
      # extend
      if not show_items:
        show_items.extend([ item.basename() for item in my.nav.nav_children ])
        if my.nav.photos and show_items:
          show_items.append('---')
        show_items.extend([ item.file.basename() for item in my.nav.photos ])
        if my.nav.files and show_items:
          show_items.append('---')
        show_items.extend([ item.basename() for item in my.nav.files ])

    return Dialog(
            None, my.dialog_sort_handler,
            dir.label(),
            my._dialog__make_tabs(
              my.dialog_sort,
              [
                di_item_list(DIALOG_ITEMCONFIG_ITEMS, 'item order',
                             show_items),
                (di_empty(), di_submit_choices( (('go', 'Go'),
                                                 ('fill', 'Fill')) ) ),
              ]
            ),
            (dir, items))

  def dialog_sort_handler(my, results, data):
    dir, prev_items = data

    if not dir == my.dir:
      return my.dialog_error(
        "Failure -- the form was sent to the wrong place??")
    dir = my.dir

    items = textarea2items(results.get(DIALOG_ITEMCONFIG_ITEMS) or '')

    if results.get('fill'):
      return my.dialog_sort(fill=items)
    
    if items != prev_items:
      sort_path = dir.child_path(SORT_NAME)

      current_items = read_entries_file(sort_path)
      if current_items != prev_items:
        print 'SORT COLLISION', current_items, 'VS', prev_items
        # darn it, someone else has edited the items in the meantime!
        return my.dialog_error(
"""Oh no! Someone else has edited the item order in the meantime!<br>
You'll have to open a new item config page and make your changes again.<br>
Sorry for the inconvenience.""")

      write_entries_file(sort_path, items)
      my.nav.refresh()

    return my.dialog_tab_choice(results, data)



  def dialog_config(my):
    dir = my.nav.dir
    
    files = my.nav.files
    for photo in my.nav.photos:
      files.append(photo.file)
    folders = my.nav.nav_children
    hidden_files = []
    hidden_folders = []

    items_hidden = my.nav.items_hidden
    items_hidden.sort()
    for name in items_hidden:
      item = dir.child(name)
      if item.is_file:
        hidden_files.append(item)
      elif item.is_dir:
        hidden_folders.append(item)

    di = []

    nr = 0
    # list all files, hidden files last.
    if len(files) + len(hidden_files) > 0:
      di.append( di_start_table('files', 'hide', 'annotate') )
      def one_file(nr, file, checked):
        return (
                 di_link(file.linkname_urlencoded(),
                         file.linkname()),
                 di_checkbox('hide_' + str(nr), '', checked=checked),
                 di_entry('annotate_' + str(nr), None,
                          my.nav.get_annotation(file), 'annotate_entry'),
               )
      for file in files:
        di.append( one_file(nr, file, False) )
        nr += 1
      for file in hidden_files:
        di.append( one_file(nr, file, True) )
        nr += 1
      di.append( di_end_table() )
      di.append( di_submit() )


    # list all folders, hidden folders last.
    if len(folders) + len(hidden_folders) > 0:
      di.append( di_start_table('folders', 'hide') )
      def one_folder(nr, folder, checked):
        return (
                 di_link(folder.linkname_urlencoded(),
                         folder.linkname()),
                 di_checkbox('hide_' + str(nr), '', checked=checked)
               )
      for folder in folders:
        di.append( one_folder(nr, folder, False) )
        nr += 1
      for folder in hidden_folders:
        di.append( one_folder(nr, folder, True) )
        nr += 1
      di.append( di_end_table() )
      di.append( di_submit() )


    return Dialog(
            my.dialog_config_handler_before,
            my.dialog_config_handler_after,
            dir.label(),
            my._dialog__make_tabs(my.dialog_config, di),
            (dir, items_hidden,
             files + folders + hidden_files + hidden_folders))




  def dialog_config_handler_before(my, results, data):
    dir, was_items_hidden, items = data

    if not dir == my.dir:
      return my.dialog_error(
        "Failure -- the form was sent to the wrong place??")
    dir = my.dir

    items_hidden_added = []
    items_hidden_removed = []

    nr = 0
    for item in items:
      name = item.basename()

      # check whether item was hidden or unhidden by the user
      if results.get('hide_' + str(nr)):
        if name not in was_items_hidden:
          items_hidden_added.append(name)
      else:
        if name in was_items_hidden:
          items_hidden_removed.append(name)

      nr += 1


    # apply the changes of items_hidden to what's currently on disk
    if items_hidden_added or items_hidden_removed:
      items_hidden_path = dir.child_path(ITEMS_HIDDEN_NAME)
      items_hidden = read_entries_file( items_hidden_path )

      changed = False
      for added in items_hidden_added:
        if added not in items_hidden:
          items_hidden.append(added)
          changed = True

      for removed in items_hidden_removed:
        if removed in items_hidden:
          items_hidden.remove(removed)
          changed = True

      if changed:
        write_entries_file(items_hidden_path, items_hidden)

    return None


  def dialog_config_handler_after(my, results, data):
    dir, was_items_hidden, items = data

    if not dir == my.dir:
      return my.dialog_error(
        "Failure -- the form was sent to the wrong place??")
    dir = my.dir

    do_refresh = False

    nr = 0
    for item in items:
      name = item.basename()

      # check whether the annotation was changed
      new_a = results.get('annotate_' + str(nr)) or ''
      was_a = item.annotation_cache or ''
      if new_a != was_a:
        # write a new annotation
        my.nav.set_annotation(item, new_a, my.trash_dir())
        do_refresh = True

      nr += 1

    if do_refresh:
      my.nav.refresh()

    return my.dialog_tab_choice(results, data)




  def dialog_copy(my):
    return my._dialog_copy_move(my.dialog_copy, 'copy', True)

  def dialog_move(my):
    return my._dialog_copy_move(my.dialog_move, 'rename / move', False)

  def _dialog_copy_move(my, choice, label, do_copy):
    dir = my.dir

    if not dir.is_root():
      di = [
             di_text(label + ' this folder to:', span=2),
             di_entry(DIALOG_ITEMCONFIG_RENAME, prefill=dir.basename()),
             di_submit(),
           ]
    else:
      di = []

    children = my.nav.files + [photo.file for photo in my.nav.photos]
    dirs = my.nav.nav_children

    for hid in my.nav.items_hidden:
      child = dir.child(hid)
      if child.is_file:
        children.append(child)
      elif child.is_dir:
        dirs.append(child)
    children.extend(dirs)

    invisible_items_at = len(children)

    # also add files and photos that are invisible but not hidden
    # explicitly.
    listing = dir.list()
    dirs = []
    for item in listing:
      if any(item == child.basename() for child in children):
        continue
      child = dir.child(item)
      if child.is_file:
        children.append(child)
      elif child.is_dir:
        dirs.append(child)
    children.extend(dirs)

    child_paths = []
    for child in children:
      child_path = '/'.join(child.rel_elements[len(dir.rel_elements):])
      child_paths.append(child_path)

    nr = 0
    if child_paths:
      if di:
        di.append( di_hr(2) )
      di.append( di_text(label + ' children to:', span=2) )

      for child_path in child_paths:
        if nr == invisible_items_at:
          di.append( di_hr(2) )
          di.append( di_text(label + ' invisible items:', style='hint') )

        di.append( 
                   (di_text(child_path, style='hint'),
                    di_entry(DIALOG_ITEMCONFIG_RENAME + "_" + str(nr), None,
                             prefill=child_path),
                   )
                 )
        nr += 1

      di.append( (di_empty(), di_submit()) )

    return Dialog(
            my.dialog_copy_move_handler, my.dialog_tab_choice,
            dir.label(),
            my._dialog__make_tabs(choice, di),
            (dir, child_paths, do_copy))

  def dialog_copy_move_handler(my, results, data):
    dir, child_paths, do_copy = data

    if not dir == my.dir:
      return my.dialog_error(
        "Failure -- the form was sent to the wrong place??")
    dir = my.dir

    def one_path(path, rename, do_copy, rel_to=None):
      if not rename or path.startswith(DB_EXLUDE):
        return None, None

      if rename.startswith('/'):
        rename_path = Path(rename.split('/'), path.base_path)
      elif rel_to:
        rename_path = rel_to.child(rename.split('/'))
      else:
        rename_path = path.sibling(rename.split('/'))

      if rename_path == path:
        rename_path = None
      elif rename_path.startswith(DB_EXLUDE):
        return ("Cannot write /" +
                path2url(rename_path.rel_path) +
                "<br>That location is protected."), rename_path
      elif rename_path.exists():
        return ("Cannot write /" +
                path2url(rename_path.rel_path) +
                "<br>That item already exists."), rename_path
      
      if rename_path:
        # move or copy it
        try:
          rename_path.make_parent_dirs()
          if do_copy:
            print "COPY", path, rename_path
            if path.is_dir:
              shutil.copytree(path.abs_path, rename_path.abs_path)
            else:
              shutil.copy2(path.abs_path, rename_path.abs_path)
          else:
            print "MOVE", path, rename_path
            shutil.move(path.abs_path, rename_path.abs_path)
        except shutil.Error:
          if do_copy:
            errname = 'copy'
          else:
            errname = 'move'
          return ("Fail -- cannot " + errname + " /" +
                  path2url(path.rel_path) + " to /" +
                  path2url(rename_path.rel_path)
                 ), rename_path
      return None, rename_path


    errors = []

    # children first
    nr = 0
    for child_path in child_paths:
      rename = results.get(DIALOG_ITEMCONFIG_RENAME + "_" + str(nr))
      if rename != child_path:
        child = dir.child(url2path(child_path))
        if child.is_file or child.is_dir:
          err, rename_path = one_path(child, rename, do_copy, dir)
      nr += 1

    if not errors:

      # rename *this* folder.
      rename = results.get(DIALOG_ITEMCONFIG_RENAME)
      if rename != dir.basename():
        err, rename_path = one_path(dir, rename, do_copy)
        if err:
          errors.append(err)
        elif rename_path:
          # move succeeded, redirect to the new URL
          my.ddd.ddd.redirect(rename_path)

    if errors:
      return my.dialog_error("<b>Errors:</b><br>" + "<br>".join(errors))

    return None

  # End of dialogs ###########################################



  def __init__(my, ddd, session, location, trailing_slash, args):
    my.ddd = ddd
    my.location = location
    my.got_trailing_slash = trailing_slash
    my.args = args
    my.s = session
    # this is the trash dir used for this request (if needed).
    # Don't use this, use trash_dir() instead.
    my._trash_dir = None
    
  def trash_dir(my):
    if not my._trash_dir:
      my._trash_dir = my._get_trash_dir()
    return my._trash_dir

  def _get_trash_dir(my):
    add = ''
    n = 0
    time_str = mtime2filename()
    while True:
      path = os.path.join(my.ddd.trash_dir.abs_path,
                          mtime2path() + add)
      if not os.path.isdir(path):
        break

      n += 1
      add = "_" + str(n)

    return path





  def render(my):
    if not my.probe_request_and_list_nav():
      return my.not_found()

    if my.file:
      if not my.file.is_file:
        return ""
      return my.ddd.ddd.serve_file(my.file.abs_path)

    nav_links = my.make_nav_links(link_to_photos=(my.do_dialog is None))

    # insert the data into the templates.
    main = 'main'
    modules = ""

    # content section
    content = ""

    if my.do_dialog:
      main = 'main_dialog'
      if my.do_dialog.has_handler():
        my.s.dialog_handler = my.do_dialog.get_handler()
      content += my.ddd.template('dialog', dialog=my.do_dialog, nav=my.nav)

    else:
      content += my.ddd.template('title', nav=my.nav)

      if my.nav.do_photo_view:
        main = 'main_photos'

        selected_photo = None
        prev_photo = None
        next_photo = None

        if len(my.nav.photos) > 0:
          selected_photo_idx = 0

          if my.selected_photo_name:
            # try to find the selected photo
            for i in range(len(my.nav.photos)):
              if my.nav.photos[i].file.basename() == my.selected_photo_name:
                selected_photo_idx = i
                break

          selected_photo = my.nav.photos[selected_photo_idx]
          if selected_photo_idx > 0:
            prev_photo = my.nav.photos[selected_photo_idx - 1]
          if selected_photo_idx < len(my.nav.photos) - 1:
            next_photo = my.nav.photos[selected_photo_idx + 1]



        content += (my.ddd.template('view_photo', nav=my.nav,
                                    photo=selected_photo,
                                    next=next_photo, prev=prev_photo)
                    + my.ddd.template('photos_thumbs', nav=my.nav,
                                      selected=selected_photo)
                   )
      else:
        content += (my.ddd.template('text', nav=my.nav)
                    + my.ddd.template('photos_thumbs', nav=my.nav,
                                      selected=None)
                    + my.ddd.template('files', nav=my.nav)
                    + my.ddd.template('child_list', nav=my.nav)
                   )

    modules += my.ddd.template('content', content=content)

    if not my.do_dialog:
      # headlet
      if not my.s.login:
        login_link = './?' + LOGIN_ARG
        logout_link = None
        add_link = None
        text_link = None
        config_link = None
      else:
        login_link = None
        logout_link = './?' + LOGOUT_ARG
        add_link = './?' + ADD_ARG
        text_link = './?' + EDIT_ARG
        config_link = './?' + CONFIG_ARG

      modules += my.ddd.template('headlet',
        text_link=text_link,
        add_link=add_link,
        config_link=config_link,
        login_link=login_link,
        logout_link=logout_link,
        last_change=mtime2str(my.nav.mtime)
        )


    # nav tree
    modules += my.ddd.template('nav', nav_links=nav_links)

    if my.nav.dir.is_root():
      title = my.nav.title
    else:
      title = my.nav.title + ' | ' + ROOT_LINK_LABEL
    return my.ddd.template(main, title=title, content=modules)


  def not_found(my):
    # find the first existing parent.
    root = Path(my.ddd.db_root)

    if not root.is_dir:
      raise "Your root path is not a directory!"

    path = root
    parent = root
    for el in my.location:
      parent = path
      path = path.child(el)

      if not path.is_dir or path.startswith(DB_EXLUDE):
        break;

    missing_elements = my.location[len(parent.rel_elements):]
    my.location = parent.rel_elements

    # make sure there will be no redirects:
    my.got_trailing_slash = True

    if len(missing_elements) < 1 or not my.probe_request_and_list_nav():
      raise "not found: " + my.location_path()

    nav_links = my.make_nav_links(do_absolute_links=True,
                                  link_to_photos=False)

    return my.ddd.template('main_dialog',
      title='not found',
      content=
        my.ddd.template('content', content=
          my.ddd.template('not_found',
            nav=my.nav,
            missing_path=my.location_path(missing_elements)) )
        +
        my.ddd.template('nav', nav_links=nav_links)
        )

  def location_path(my, elements=None):
    return '/'.join(elements or my.location)


  def probe_request_and_list_nav(my):
    my.error = None
    my.file = None
    my.dir = None
    my.nav = None
    my.selected_photo_name = None
    my.do_photo_view = False
    my.do_dialog = None

    # check the requested path itself. Is it a file? is it a dir?
    # is it supposed to be on the web?
    root = Path(my.ddd.db_root)
    path = root.child(my.location)

    if path.startswith(DB_EXLUDE):
      return False

    if PHOTO_VIEW_ARG in my.args:
      # show the photo gallery.
      my.do_photo_view = True

      # remember which photo is selected to view.
      # 
      # This is the counterpart to Photo.view_url(). The selected
      # photo's name is always attached as a simple child filename, even if
      # the real file is in "_photos/".
      if not path.is_dir:
        my.selected_photo_name = path.basename()
        # Do the nav listing of this photo's parent folder.
        path = path.parent()

    elif path.is_file:
      # Just serve the file, don't even consider anything else.
      my.file = path
      return True

    elif path.is_dir:
      # to enable relative urls when a requested folder url has no trailing
      # slash, we need to redirect to the same url *with* a trailing slash.
      # cherrypy doesn't do this for us since we manage the paths ourselves.
      if not my.got_trailing_slash:
        print "REDIRECT to add a slash", '/' + path2url(path.rel_path)
        my.ddd.ddd.redirect(path)

    # At this point, we might be given a dir to process instead of an
    # original file path

    if not path.is_dir:
      # it's either missing or we don't handle it.
      return False
    
    my.dir = path

    # Handle GET and POST request arguments (1 of 2)
    # ---------------------------------------------------
    # (Note, PHOTO_VIEW_ARG and LO_RES_ARG are handled above)

    # handle attribute-only requests
    if LOGOUT_ARG in my.args:
      # logging out.
      my.s.login = None

    # handle any submits of previously shown dialogs
    if my.s.dialog_handler:
      my.do_dialog = my.s.dialog_handler.handle(my, my.args,
                                                is_before_nav=True)
    
    # ---------------------------------------------------


    # list my children
    my.nav = DirNav(my, do_photo_view=my.do_photo_view)


    # Handle GET and POST request arguments (2 of 2)
    # ---------------------------------------------------
    
    # handle any submits of previously shown dialogs
    if my.s.dialog_handler:
      dialog = my.s.dialog_handler.handle(my, my.args, is_before_nav=False)
      if dialog:
        my.do_dialog = dialog

    if not my.do_dialog:
      # launch new dialogs if requested
      if LOGIN_ARG in my.args:
        my.do_dialog = my.dialog_login()

      if my.s.login:
        if ADD_ARG in my.args:
          my.do_dialog = my.dialog_add()
        elif REMOVE_ARG in my.args:
          my.do_dialog = my.dialog_remove()
        elif EDIT_ARG in my.args:
          my.do_dialog = my.dialog_text()
        elif CONFIG_ARG in my.args:
          my.do_dialog = my.dialog_last_tab_choice()
    
    # ---------------------------------------------------

    return True



  def make_nav_links(my, do_absolute_links=False, link_to_photos=True):

    class NavLink:
      def __init__(my, link, label, style):
        "make sure LINK is urlencoded!"
        my.link = link
        my.label = unicode(label, 'utf-8').encode("ascii", "xmlcharrefreplace")
        my.style = style

      def __str__(my):
        #return '(' + my.link + ', ' + my.label + ', ' + my.style + ')'
        return my.label


    class NavLinkCascade:
      def __init__(my, *root_navlink_args):
        my.root_navlink_args = root_navlink_args
        my.links = []
        my.levels = 1
        my.deeper_items = None
        my.parent_of_deeper_items = None

      def append(my, *args):
        my.links.append(NavLink(*args))

      def prepend(my, *args):
        my.links.insert(0, NavLink(*args))

      def wrap(my, *parent_args):
        """wraps the link list in another list, in effect opening up
        a new nav level. Keeps count."""

        # re-insert previous deeper item cascade, if one is still around.
        # Then it wasn't one of the items in the list (probably hidden),
        # just list it on the top.
        my.prepend_deeper_items()

        if ( len(my.links) == 0 or
             len(my.links) == 1 and type(my.links[0]) is list ):
           # don't wrap nothing.
           my.parent_of_deeper_items = parent_args
           return

        # save deeper stuff
        # will be re-inserted by my.result() or my.append_deeper_items()
        my.parent_of_deeper_items = parent_args
        my.deeper_items = my.links

        # new outer list. Things will be gradually appended to the outer
        # list, and as soon as an item matches the deeper list, the deeper
        # list is inserted at that position. If the deeper list was not
        # encountered (e.g. for an item hidden from the nav), the deeper
        # list will be prepended to the start.
        my.links = []
        my.levels += 1

      def append_deeper_items(my):
        if my.parent_of_deeper_items:
          my.append(*my.parent_of_deeper_items)
        if my.deeper_items:
          my.links.append(my.deeper_items)
        my.deeper_items = None
        my.parent_of_deeper_items = None

      def prepend_deeper_items(my):
        if my.deeper_items:
          my.links.insert(0, my.deeper_items )
        if my.parent_of_deeper_items:
          my.prepend(*my.parent_of_deeper_items)
        my.deeper_items = None
        my.parent_of_deeper_items = None

      def result(my):
        my.wrap(*my.root_navlink_args)
        my.append_deeper_items()
        return my.links

      def str(my, links):
        r = '['
        for item in links:
          if type(item) is list:
            r += my.str(item) + ', '
          else:
            r += str(item) + ', '
        r += ']'
        return r

    # end of NavLinkCascade


    root_url = my.ddd.root_url

    if my.nav.dir.is_root():
      root_style = 'root_selected'
    else:
      root_style = 'root'

    # The links listing in nested lists, each nested list causing more
    # indent. We are constructing from the deepest level outwards.
    cascade = NavLinkCascade(root_url, ROOT_LINK_LABEL, root_style)

    dir = my.nav.dir
    nav = my.nav
    child_nav = None

    # the links and styles for the first loop iteration
    link_base = ''

    if do_absolute_links:
      link_base = my.nav.dir.abs_url(my.ddd.root_url)
      if not link_base.endswith('/'):
        link_base += '/'

    child_style = 'child'
    if not nav.do_photo_view and link_to_photos:
      link = link_base + '?' + PHOTO_VIEW_ARG
    else:
      link = link_base + './'
    style = 'selected'

    while True:

      if nav:
        # add my children
        for child in nav.nav_children:
          if child_nav and child == child_nav.dir:
            cascade.append_deeper_items()
            # skip adding the already added child_nav
            continue

          cascade.append(link_base + child.linkname_urlencoded(),
                         child.basename(),
                         child_style)


      if dir.is_root():
        return cascade.result()

      cascade.wrap(link, dir.basename(), style)


      # Add the siblings of the current directory.
      # So, get a listing from the current directory's parent.
      child_nav = nav
      nav = None
      dir = dir.parent()
      if cascade.levels <= NAV_MAX_EXPANDED_LEVELS:
        nav = DirNav(my, dir, nav_only=True)
      if do_absolute_links:
        link_base = dir.abs_url(my.ddd.root_url)
      else:
        link_base += '../'
      # And actually, adding them will be done by the next loop iteration.


      if cascade.levels >= NAV_MAX_LEVELS:
        # at this point, the whole cascade has just been wrapped and is
        # waiting in the background to be inserted somewhere. But we are
        # now at the outermost allowed level.
        if not dir.is_root():
          # there already is a root link. For all others that are
          # cut off, show a '...' link to lead upwards.
          cascade.append(link_base, '...', 'more_ancestors')
          # append the deeper items, so they are not *pre*pended by
          # cascade.result().
          cascade.append_deeper_items()
        return cascade.result()
      

      # Set the values for the next loop iterations
      link = link_base
      style = 'ancestor'
      if child_style == 'child':
        child_style = 'sibling' 
      else:
        child_style = 'cousin'





def config_abs_path(config_path, relative_to):
  if os.path.isabs(config_path):
    return config_path
  return os.path.join(relative_to, config_path)
    


class _Ddd:
  """This is the API to ddd, called from exposed functions. It's job
  is to read and store the config, manage sessions and launch a new
  handler object for each request.
  It is split off to cleanly separate from cherrypy."""

  def __init__(my, ddd, ddd_location):
    """Determine any automated config, store in this persistent object."""
    print "NEW _Ddd"
    my.ddd = ddd

    my.script_location = ddd_location

    my.db_root = DB_FILE_SYSTEM_LOCATION
    if my.db_root is None:
      my.db_root = os.path.dirname(my.script_location)
    print "_Ddd: db_root =", my.db_root

    if os.path.isabs(TRASH_PATH):
      trash_path_abs = os.path.abspath(TRASH_PATH)
      if trash_path_abs.startswith(my.db_root):
        my.trash_dir = Path(trash_path_abs, my.db_root)
      else:
        my.trash_dir = Path(trash_path_abs)
      print '_Ddd: 2trash_dir =', my.trash_dir
    else:
      my.trash_dir = Path(TRASH_PATH, my.db_root)
    print '_Ddd: trash_dir =', my.trash_dir

    my.thumbs_dir = Path(THUMBS_DIR, my.db_root)
    my.lo_res_dir = Path(LO_RES_DIR, my.db_root)

    my.templates_dir = config_abs_path(TEMPLATES_DIR, my.script_location)
    print "_Ddd: templates_dir =", my.templates_dir

    my.templates = {}

    my.root_url = ROOT_URL or '/'
    if not my.root_url.endswith('/'):
      my.root_url += '/'

    print "_Ddd: root_url =", my.root_url

      
  def handle(my, s, location, trailing_slash, args):
    return DddRequestHandler(my, s, location, trailing_slash, args).render()
    
  def template(my, name, **props):
    # This is all we need from Cheetah:
    from Cheetah.Template import Template
    file = Path(os.path.join(my.templates_dir, name))
    template = None
    if name in my.templates:
      template, mtime = my.templates[name]
      if mtime < file.mtime():
        template = None
    
    if not template:
      print "COMPILING TEMPLATE", name
      mtime = file.mtime() # (read the mtime before reading the file)
      template = Template.compile(file=file.abs_path)
      my.templates[name] = (template, mtime)

    root = my.root_url
    if root.endswith('/'):
      root = root[:-1]
    props['root'] = root
    props['root_label'] = ROOT_LINK_LABEL
      
    return str(template(searchList=props))
 
  def is_trash_or_image_cache(my, path):
    "PATH should be a Path object."
    return my.is_trash(path) or my.is_image_cache(path)

  def is_trash(my, path):
    "PATH should be a Path object."
    print 'IS TRASH?', path.abs_path, my.trash_dir.abs_path
    return path.abs_path.startswith(my.trash_dir.abs_path)

  def is_image_cache(my, path):
    "PATH should be a Path object."
    print 'IS IMAGE CACHE?', path.abs_path, my.thumbs_dir.abs_path, my.lo_res_dir.abs_path
    return ( path.abs_path.startswith(my.thumbs_dir.abs_path) or
             path.abs_path.startswith(my.lo_res_dir.abs_path) )

  def exists(my, url):
    while url.startswith('/'):
      url = url[1:]
    return os.path.exists(os.path.join(my.db_root, url2path(url)))

# This is all we need from cherrypy:
import cherrypy
from cherrypy.lib.static import serve_file
from cherrypy import _cperror

def handle_error():
    cherrypy.response.status = 500
    cherrypy.response.body = ['''<h2>Internal error</h2>
Sorry, something went wrong.<br><br>
<a href="''' + (ROOT_URL or '/') + '">' + ROOT_LINK_LABEL + '</a>']
    print '***  ERROR:\n', _cperror.format_exc()



class Session:
  def __init__(s):
    s.login = None
    s.dialog_handler = None
    s.last_dialog_tab = None


class Ddd:
  exposed = True
  _cp_config = {'request.error_response': handle_error}

  def __init__(self, dir, do_test=None):
    self.ddd = _Ddd(self, dir)
    self.do_test = do_test


  def __call__(my, *location, **args):
    print
    print "-------------------------------"

    if my.do_test:
      session = Session()
      if my.do_test[2]:
        session.login = 'user'
    else:
      key = 'ddd_session'
      if 'kill' in args:
        session = None
      else:
        session = cherrypy.session.get(key)
      if not session:
        session = Session()
        cherrypy.session[key] = session

    if my.do_test:
      url = my.do_test[0]
      # and use the supplied LOCATION from above.
    else:
      # cherrypy tries to be nice and parses the request URL.
      # However, we can't use that :(
      # There are these issues:
      # (1) trailing slash
      #     We need to know if a trailing slash is there or not. If
      #     the requested item is a folder, we may need to redirect to
      #     add a trailing slash, or all relative URLs will break.
      # (2) question mark in item name
      #     If the name of a folder contains a question mark, cherrypy
      #     takes the question mark as the start of the arguments.
      #       localhost:8080/one?two/?arg=value
      #       --> "/one" not found.
      # If you know something better, tell me.

      try:
        # parse the GET or POST header line, allow spaces in the URL part.
        url = ' '.join(cherrypy.request.request_line.split(' ')[1:-1])
        
        tokens = url.split('?')
        print 'TOKENS', tokens
        
        if len(tokens) > 1:
          arglist = tokens[-1]
          print 'ARGLIST', arglist
          # ddd never issues a slash in an ?arg
          if arglist.find('/') < 0 and not my.ddd.exists(url):
            # there's no slash after the '?', and the full path including
            # the '?' does not exist. Parse the arguments.
            url = '?'.join(tokens[:-1])
            # Cherrypy already parsed the arguments, and also put POST
            # args in there.
            # However, make sure the one we see is the real first arg is
            # actually added. ddd only uses a single GET arg without value.
            args[arglist] = ''

      except:
        url = "/"

      location = [el for el in url.split('/') if el]

    print 'URL', url, args
    print 'LOC', location

    return my.ddd.handle(session, location, url.endswith('/'), args)


  def serve_file(my, *args):
    return cherrypy.lib.static.serve_file(*args)

  def redirect(my, path):
    if my.do_test:
      print "IGNORED: Redirect to", path.rel_path
    else:
      url = cherrypy.url('/') + urllib.quote(path2url(path.rel_path))
      print 'REDIRECT TO', url
      if path.is_dir:
        url += '/'
      raise cherrypy.HTTPRedirect(url)


def where_am_i():
  path = os.path.dirname(sys.argv[0])
  if not path.startswith(os.path.sep):
    path = os.path.join( os.getcwd(), path )
  return os.path.abspath(path)

  

dir = where_am_i()



# parse arguments
import getopt
optlist, args = getopt.getopt(sys.argv[1:], 'p:tl', ['pid-file', 'test'])

do_test = None
test_logged_in = False

for opt, opt_arg in optlist:
  if opt == '-t':
    do_test = True

  elif opt == '-p':
    pid_file = Path(opt_arg)
    pid_file.rewrite(str(os.getpid()))
  
  elif opt == '-l':
    test_logged_in = True
  
if do_test:
  # if run with an arg, test it from commandline
  if len(args) > 0:
    dict = {}
    print 'ARGS', args[1:]
    for arg in args[1:]:
      name, value = arg.split('=')
      dict[name] = value
    do_test = [args[0], dict, test_logged_in]
  else:
    do_test = ['/', {}, test_logged_in]

  url = [el for el in do_test[0].split('/') if len(el) > 0]
  print "Testing URL:", '/' + '/'.join(url), '?', do_test[1]
  print Ddd(dir, do_test=do_test)(*url, **(do_test[1]))
  exit(0)
else:
  if test_logged_in:
    print "-l only works with -t"
    exit(1)

favicon = config_abs_path(FAVICON_PATH, dir)

print "FAVICON_PATH", favicon

cherrypy.config.update({'server.socket_port': DDD_PORT})

root_config = {
               'tools.sessions.on': True,
               'tools.sessions.timeout': SESSIONS_TIMEOUT,
              }

if SESSIONS_DIR:
  sessions_dir = config_abs_path(SESSIONS_DIR, dir)
  root_config['tools.sessions.storage_type'] = "file"
  root_config['tools.sessions.storage_path'] = sessions_dir
  if not os.path.isdir(sessions_dir):
    os.makedirs(sessions_dir)
    os.chmod(sessions_dir, stat.S_IREAD | stat.S_IWRITE |
                           stat.S_IRWXU)
  print "SESSIONS_DIR", sessions_dir

config = {'/': root_config,
          '/favicon.ico': {
            'tools.staticfile.on': True,
            'tools.staticfile.filename': favicon,
          }

         }

cherrypy.quickstart(Ddd(dir), config=config)

